/* Author: Domagoj Babic 2004.
 *
 * $Id: graph.h,v 1.5 2005/02/21 22:32:30 root Exp $
 *
 */

/***********************************************************************
*
*  Copyright (C) 2004-2005, by Domagoj Babic (Author). All rights 
*  reserved. By using this software the USER indicates that he or she 
*  has read, understood and will comply with the following:
*
*  This software may not be called "HyperSAT" if it has been modified in 
*  any way, without the specific prior written permission of the Author.
*
*  Author hereby grants USER nonexclusive permission to use, copy and/or 
*  modify this software for internal, noncommercial, research purposes 
*  only. Any distribution, including commercial sale or license, of this 
*  software, copies of the software, its associated documentation and/or 
*  modifications of either is strictly prohibited without the prior 
*  consent from the Author. Appropriate copyright notice shall be placed 
*  on all software copies, and a complete copy of this notice shall be 
*  included in all copies of the associated documentation.
*
*  THE SOFTWARE IS PROVIDED "AS-IS" AND AUTHOR MAKES NO REPRESENTATIONS 
*  OR WARRANTIES, EXPRESS OR IMPLIED, BY WAY OF EXAMPLE, BUT NOT 
*  LIMITATION. AUTHOR MAKES NO REPRESENTATIONS OR WARRANTIES OF 
*  MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR PURPOSE OR THAT THE
*  USE OF THE SOFTWARE WILL NOT INFRINGE ANY PATENTS, COPYRIGHTS
*  TRADEMARKS OR OTHER RIGHTS. AUTHOR SHALL NOT BE LIABLE FOR ANY
*  LIABILITY OR DAMAGES WITH RESPECT TO ANY CLAIM BY LICENSEE OR ANY
*  THIRD PARTY ON ACCOUNT OF, OR ARISING FROM THE LICENSE, OR ANY
*  SUBLICENSE OR USE OF THE SOFTWARE OR ANY SERVICE OR SUPPORT.
*
*  Author shall not be liable under any circumstances for any direct, 
*  indirect, special, incidental, or consequential damages with respect 
*  to any claim by USER or any third party on account of or arising from 
*  the use, or inability to use, this software or its associated 
*  documentation, even if the Author has been advised of the possibility 
*  of those damages.
*
*  The above copyright notice must appear in all copies of the software.
*
************************************************************************
*
*  Support notice: Unfortunatelly, the author can't guarantee any 
*  support. The solver has been written exclusively for research 
*  and experimental purposes, so documentation and comments are scarce.
*  Most likely, this will not change in the near future, but feel free
*  to check my webpage occasionally, as new releases (if there will be
*  any) will be made available at my website:
*  http:/www.cs.ubc.ca/~babic/
*
***********************************************************************/

#include <stdlib.h>
#include "utils.h"
#include "queues.h"

#ifndef H_GRAPH
#define H_GRAPH

enum vertex_flags_e {VERTEX_ROOT_EQV = 1, VERTEX_MARK = 2};

// graph node
typedef struct gnd_t {
    uint v;     // vertex
    struct gnd_t *nxt;
    uint eqv;   // equivalency relation
} gnd_t;

// adjacency list representation of graph
typedef struct al_grph_t {
    uint vxnum;     // number of vertices
    uint *tmp;      // temporary array for passing search results
    stack_t *vex;   // added equivalence vertices
    stack_t *s;     // auxiliary stack
    gnd_t **mtx;    // node ptr matrix
    char *nflags;   // node flags
} al_grph_t;

inline char grph_is_marked(al_grph_t *g, uint v) {
    assert(g != NULL);
    return g->nflags[v] & VERTEX_MARK;
}

inline void grph_mark(al_grph_t *g, uint v) {
    assert(g != NULL);
    assert(!grph_is_marked(g, v));
    g->nflags[v] |= VERTEX_MARK;
}

inline void grph_del_mark(al_grph_t *g, uint v) {
    assert(g != NULL);
    assert(grph_is_marked(g, v));
    g->nflags[v] &= ~VERTEX_MARK;
}

inline uint grph_get_root_eqv(al_grph_t *g, uint v) {
    assert(g != NULL);
    return (g->nflags[v] & VERTEX_ROOT_EQV) ? TRUE : FALSE;
}

inline void grph_set_root_eqv(al_grph_t *g, uint v, uint eqv) {
#ifdef UGLY_HACKS
#ifdef DEBUG
    char t, ta;
#endif
#endif
    assert(g != NULL);
    assert(v < g->vxnum);
    assert(eqv < 2);
#ifdef UGLY_HACKS
#ifdef DEBUG
    t = g->nflags[v];
    g->nflags[v] ^= (-eqv ^ g->nflags[v]) & VERTEX_ROOT_EQV;
    ta = g->nflags[v];
    assert(!eqv || ta == (t | VERTEX_ROOT_EQV));
    assert(eqv || ta == (t & ~VERTEX_ROOT_EQV));
#else
    g->nflags[v] ^= (-eqv ^ g->nflags[v]) & VERTEX_ROOT_EQV;
#endif
#else
    g->nflags[v] = (eqv) ? g->nflags[v] | VERTEX_ROOT_EQV :
        g->nflags[v] & ~VERTEX_ROOT_EQV;
#endif
}

gnd_t *grph_newnode(uint v) {
    gnd_t *n = xcalloc(1, sizeof(gnd_t));
    n->v = v;
    return n;
}

al_grph_t *grph_new_al_grph(uint size) {
    al_grph_t *g = xcalloc(1, sizeof(al_grph_t));
    g->vxnum = size + 1;
    g->nflags = xcalloc(1, g->vxnum);
    g->tmp = xcalloc(1, g->vxnum * sizeof(uint)); 
    g->vex = initStack(16);
    g->s = initStack(16);
    g->mtx = (gnd_t **)xcalloc(1, g->vxnum * sizeof(gnd_t *));
    return g;
}

void free_al_grph(al_grph_t *g) {
    uint i;
    for (i = 0; i < g->vxnum; i++) {
        gnd_t *n = g->mtx[i];
        while (n) {
            gnd_t *t = n->nxt;
            free(n);
            n = t;
        }
    }
    free(g->tmp);
    free(g->mtx);
    free(g->nflags);
    stackFree(g->vex);
    stackFree(g->s);
    free(g);
}

void grph_ins_nd(al_grph_t *g, uint v1, uint v2, uint eqv) {
    gnd_t *n = grph_newnode(v2);
    assert(!grph_is_marked(g, v1));
    assert(!grph_is_marked(g, v2));
    n->nxt = g->mtx[v1];
    n->eqv = eqv;
    g->mtx[v1] = n;
}

// eqv = 1 if v1 <==> v2 (v1\/-v2)(-v1\/v2)
// eqv = 0 if v1 <!=> v2 (-v1\/-v2)(v1\/v2)
void grph_eqv_insert(al_grph_t *g, uint v1, uint v2, uint eqv) {
    assert(eqv == TRUE || eqv == FALSE);
    assert(v1 < g->vxnum && v2 < g->vxnum);
    //printf("Inserting bin eq: (%u, %u):%u\n", v1, v2, eqv);
    grph_ins_nd(g, v1, v2, eqv);
    grph_ins_nd(g, v2, v1, eqv);
    stackPush(g->vex, v1);
    stackPush(g->vex, v2);
}

// root_eqv = 1 if v <=> root_node
uint grph_push_nds(al_grph_t *g, uint v, uint root_eqv) {
    gnd_t *n = g->mtx[v];
    assert(v < g->vxnum);
    assert(root_eqv == TRUE || root_eqv == FALSE);
    while (n) {
        uint x = n->v;
        // n->eqv == 1 if x <=> v
        uint eqv = !(n->eqv ^ root_eqv);
        if (!grph_is_marked(g, x)) {
            stackPush(g->s, (Item)eqv);
            stackPush(g->s, (Item)x);
        } else if (eqv != grph_get_root_eqv(g, x)) {
            return FALSE;
        }
        n = n->nxt;
    }
    return TRUE;
}

// returns TRUE if there are any equivalences 
inline uint grph_any_eqs(al_grph_t *g) {
    return !stackEmpty(g->vex);
}

// alternating depth first search
uint *grph_alt_dfs(al_grph_t *g) {
    uint v, i = 0;
#ifdef DEBUG
    uint j, x = 0, sz, *sorted;
#endif
    assert(g != NULL);
    assert(grph_any_eqs(g));
    do {
        v = stackPop(g->vex);
    } while (grph_any_eqs(g) && grph_is_marked(g, v));

    assert(v < g->vxnum);
    // nodes might be pushed twice to the stack before beeing
    // marked, so when such a variable is marked, the algorithm
    // will run into it again later. as each variable needs to
    // be visited only once, the algorithm has to skip marked
    // vars. also, the root variable could be pushed twice to
    // the stack, and found after beeing unmarked in second
    // iteration. at that point, root variable will not have
    // any nodes associated to it, and it needs to be skipped.
    if (!grph_is_marked(g, v) && g->mtx[v]) {
        stackPush(g->s, (Item)TRUE); // root var is equivalent to itself
        stackPush(g->s, (Item)v);

        while (!stackEmpty(g->s)) {
            uint v = stackPop(g->s);             // vertex
            uint root_eqv = (uint)stackPop(g->s);// root equivalency
            if (grph_is_marked(g, v)) {
                // the same node could have been pushed to stack
                // twice before beeing visited. in that case
                // we might pop up a marked node from the stack.
#ifdef DEBUG
                // here we check that that variable has actually
                // been added to tmp array in this run
                char found = FALSE;
                uint k;
                for (k = 0; k < i; k++) {
                    if (g->tmp[k] >> 1 == v) found = TRUE;
                }
                assert(found);
#endif
                continue;
            }
            grph_mark(g, v);
            grph_set_root_eqv(g, v, root_eqv);
            assert(root_eqv == TRUE || root_eqv == FALSE);
            // root_eqv = 1 if v <==> rootv
            g->tmp[i++] = (v << 1) | root_eqv;
            if (!grph_push_nds(g, v, root_eqv)) {
                return NULL;
            }
        }    
    }
    g->tmp[i] = 0; // sentinel
#ifdef DEBUG
    sz = i + 1;
    sorted = xmalloc(sz * sizeof(uint));
    memcpy(sorted, g->tmp, sz * sizeof(uint));
    cmp_qsort(sorted, sz, 4, ucmp);
    for (j = 0; j < sz; j++) {
        if (j > 0) {
            assert(x != sorted[j]);
        }    
        x = sorted[j];
    }
    free(sorted);
#endif
    return g->tmp;        
}

void grph_free_children(al_grph_t *g, uint v) {
    gnd_t *n = g->mtx[v];
    stack_t *s = g->s;
    while (n) {
        gnd_t *t = n->nxt;
        stackPush(s, n->v);
        free(n);
        n = t;
    }
    g->mtx[v] = NULL;
    while (!stackEmpty(s)) {
        grph_free_children(g, stackPop(s));
    }
}

void grph_reset_root(al_grph_t *g, uint rootv) {
    grph_free_children(g, rootv);
    grph_del_mark(g, rootv);
}
#endif

/*
 * Revision history:
 *
 * $Log: graph.h,v $
 * Revision 1.5  2005/02/21 22:32:30  root
 * SAT competition release.
 *
 * Revision 1.4  2004/12/23 23:07:34  root
 * Optimized B-cubing implementation (now the algorithm picks the
 * variable with the highest priority as the new branch node and
 * search backtracks over unused BCT branch nodes). Fixed some
 * minor bugs. Code passes splint -weak check.
 *
 * Revision 1.2  2004/12/09 18:48:27  root
 * Minor change.
 *
 * Revision 1.1  2004/10/12 23:31:15  root
 * Initial import.
 *
 *
 */
